import numpy as np
from matplotlib import animation, rc
import matplotlib.pyplot as plt
import tonic.transforms as transforms


def pad_events(batch):
    max_length = 0
    for sample, target in batch:
        if len(sample) > max_length:
            max_length = len(sample)
    samples_output = []
    targets_output = []
    for sample, target in batch:
        sample = np.vstack((np.zeros((max_length - len(sample), 4)), sample))
        samples_output.append(sample)
        targets_output.append(target)
    return np.stack(samples_output), targets_output


# needs matplotlib widget backend
def plot_events(events, sensor_size, ordering, frame_time=25000, repeat=False):
    from .utils import plot_frames

    transform = transforms.Compose(
        [
            transforms.MaskIsolated(time_filter=20000),
            transforms.ToRatecodedFrame(frame_time=frame_time, merge_polarities=True),
        ]
    )
    frames = transform(events, sensor_size=sensor_size, ordering=ordering)
    plot_frames(frames, frame_time)


def plot_frames(frames, frame_time, repeat=False, vmax=255):
    plt.close()
    fig, ax = plt.subplots()
    axisImage = ax.imshow(frames[0, :, :], cmap=plt.cm.plasma, vmin=0, vmax=vmax)

    def animate(i):
        fig.suptitle("frames {0}".format(i))
        axisImage.set_data(frames[i, :, :])
        return (axisImage,)

    return animation.FuncAnimation(
        fig,
        animate,
        frames=frames.shape[0],
        interval=frame_time / 1000,
        blit=True,
        repeat=repeat,
    )



from .crop import crop_numpy
from .drop_events import drop_events_numpy
from .flip_lr import flip_lr_numpy
from .flip_polarity import flip_polarity_numpy
from .flip_ud import flip_ud_numpy
from .denoise import denoise_numpy
from .mix_ev_streams import mix_ev_streams_numpy
from .refractory_period import refractory_period_numpy
from .spatial_jitter import spatial_jitter_numpy
from .st_transform import st_transform
from .time_jitter import time_jitter_numpy
from .time_reversal import time_reversal_numpy
from .time_skew import time_skew_numpy
from .to_averaged_timesurface import to_averaged_timesurface
from .to_ratecoded_frame import to_ratecoded_frame_numpy
from .to_sparse_tensor import to_sparse_tensor_pytorch
from .to_timesurface import to_timesurface_numpy
from .uniform_noise import uniform_noise_numpy
from .utils import guess_event_ordering_numpy, is_multi_image

__all__ = [
    "crop_numpy",
    "drop_events_numpy",
    "flip_lr_numpy",
    "flip_polarity_numpy",
    "flip_ud_numpy",
    "denoise",
    "mix_ev_streams_numpy",
    "refractory_period_numpy",
    "spatial_jitter_numpy",
    "st_transform",
    "time_jitter_numpy",
    "time_reversal_numpy",
    "time_skew_numpy",
    "to_averaged_timesurface",
    "to_ratecoded_frame_numpy",
    "to_sparse_tensor_pytorch",
    "to_timesurface_numpy",
    "uniform_noise_numpy",
]



import os
import numpy as np
from torchvision.datasets.vision import VisionDataset
from torchvision.datasets.utils import (
    check_integrity,
    download_and_extract_archive,
    extract_archive,
)


class POKERDVS(VisionDataset):
    """POKER DVS <http://www2.imse-cnm.csic.es/caviar/POKERDVS.html> data set

    arguments:
        save_to: location to save files to on disk
        train: choose training or test set
        download: choose to download data or not
        transform: list of transforms to apply to the data
        target_transform: list of transforms to apply to targets
    """

    base_url = "https://www.neuromorphic-vision.com/public/downloads/"
    train_filename = "pips_train.tar.gz"
    test_filename = "pips_test.tar.gz"
    train_url = base_url + train_filename
    test_url = base_url + test_filename
    train_md5 = "412bcfb96826e4fcb290558e8c150aae"
    test_md5 = "eef2bf7d0d3defae89a6fa98b07c17af"

    classes = ["cl", "he", "di", "sp"]
    int_classes = dict(zip(classes, range(4)))
    sensor_size = (35, 35)
    ordering = "txyp"

    def __init__(
        self, save_to, train=True, download=True, transform=None, target_transform=None
    ):
        super(POKERDVS, self).__init__(
            save_to, transform=transform, target_transform=target_transform
        )

        self.train = train
        self.location_on_system = save_to
        self.data = []
        self.targets = []

        if train:
            self.url = self.train_url
            self.file_md5 = self.train_md5
            self.filename = self.train_filename
            self.folder_name = "pips_train"
        else:
            self.url = self.test_url
            self.file_md5 = self.test_md5
            self.filename = self.test_filename
            self.folder_name = "pips_test"

        if download:
            self.download()

        if not check_integrity(
            os.path.join(self.location_on_system, self.filename), self.file_md5
        ):
            raise RuntimeError(
                "Dataset not found or corrupted."
                + " You can use download=True to download it"
            )

        file_path = self.location_on_system + "/" + self.folder_name
        for path, dirs, files in os.walk(file_path):
            dirs.sort()
            for file in files:
                if file.endswith("npy"):
                    self.data.append(np.load(path + "/" + file))
                    self.targets.append(self.int_classes[path[-2:]])

    def __getitem__(self, index):
        events, target = self.data[index], self.targets[index]
        if self.transform is not None:
            events = self.transform(events, self.sensor_size, self.ordering)
        if self.target_transform is not None:
            target = self.target_transform(target)
        return events.astype("int64"), target

    def __len__(self):
        return len(self.data)

    def download(self):
        download_and_extract_archive(
            self.url, self.location_on_system, filename=self.filename, md5=self.file_md5
        )



import numpy as np
import torch


def to_sparse_tensor_pytorch(
    events, sensor_size, ordering, merge_polarities=False,
):
    """Sparse Tensor PyTorch representation. See https://pytorch.org/docs/stable/sparse.html for details.

    Args:
        merge_polarities (bool): flag that tells whether polarities should be taken into account separately or not.

    Returns:
        sparse tensor in TxWxH format
    """
    assert "x" and "y" and "t" and "p" in ordering
    x_index = ordering.find("x")
    y_index = ordering.find("y")
    t_index = ordering.find("t")
    p_index = ordering.find("p")

    if len(events.shape) != 2:
        raise RuntimeError(
            "Will only convert to sparse tensor from (N,E) (i.e., a list of events)"
            " dimension."
        )

    if merge_polarities:
        events[:, p_index] = np.zeros(n_of_events)
    else:
        pols = events[:, p_index]
        pols[pols == 0] = -1

    max_time = int(max(events[:, t_index]) + 1)
    max_x = int(max(events[:, x_index]) + 1)
    max_y = int(max(events[:, y_index]) + 1)

    indices = torch.LongTensor(events[:, [t_index, x_index, y_index]]).T
    values = torch.FloatTensor(events[:, p_index])
    return torch.sparse.FloatTensor(
        indices, values, torch.Size([max_time, max_x, max_y,])
    )



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
import sphinx_rtd_theme

sys.path.insert(0, os.path.abspath(".."))

# -- Project information -----------------------------------------------------

project = "Tonic"
copyright = "2020, the Neuromorphs"
author = "the Neuromorphs"

try:
    import tonic

    version = ".".join(tonic.__version__.split(".")[:2])
    release = tonic.__version__
except ImportError:
    version = ""
    release = ""

master_doc = "index"

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ["sphinx.ext.autodoc", "sphinx.ext.napoleon", "sphinx_rtd_theme"]

# Napoleon settings
napoleon_google_docstring = True
napoleon_numpy_docstring = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.

html_theme = "sphinx_rtd_theme"
html_logo = "_static/tonic-logo-white.png"
html_show_sourcelink = True

html_theme_options = {"logo_only": True, "analytics_id": "UA-36748283-2"}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]



import numpy as np

from .utils import guess_event_ordering_numpy


def spatial_jitter_numpy(
    events,
    sensor_size,
    ordering=None,
    variance_x=1,
    variance_y=1,
    sigma_x_y=0,
    integer_coordinates=True,
    clip_outliers=True,
):
    """Changes position for each pixel by drawing samples from a multivariate
    Gaussian distribution with the following properties:

        mean = [x,y]
        covariance matrix = [[variance_x, sigma_x_y],[sigma_x_y, variance_y]]

    Jittered events that lie outside the focal plane will be dropped if clip_outliers is True.

    Args:
        events: ndarray of shape [num_events, num_event_channels]
        ordering: ordering of the event tuple inside of events, if None
                  the system will take a guess through
                  guess_event_ordering_numpy. This function requires 'x'
                  and 'y' to be in the ordering
        variance_x: squared sigma value for the distribution in the x direction
        variance_y: squared sigma value for the distribution in the y direction
        sigma_x_y: changes skewness of distribution, only change if you want shifts along diagonal axis.
        integer_coordinates: when True, shifted x and y values will be integer coordinates
        clip_outliers: when True, events that have been jittered outside the focal plane will be dropped.

    Returns:
        spatially jittered set of events.
    """

    if ordering is None:
        ordering = guess_event_ordering_numpy(events)
    assert "x" and "y" in ordering

    x_index = ordering.find("x")
    y_index = ordering.find("y")

    shifts = np.random.multivariate_normal(
        [0, 0], [[variance_x, sigma_x_y], [sigma_x_y, variance_y]], len(events)
    )

    if integer_coordinates:
        shifts = shifts.round()

    xs = events[:, x_index]
    ys = events[:, y_index]

    if np.issubdtype(events.dtype, np.integer):
        xs += shifts[:, 0].astype(np.int)
        ys += shifts[:, 1].astype(np.int)
    else:
        xs += shifts[:, 0]
        ys += shifts[:, 1]

    if clip_outliers:
        events = np.delete(events, (np.where((xs < 0) | (xs > sensor_size[0]))), axis=0)
        events = np.delete(events, (np.where((ys < 0) | (ys > sensor_size[1]))), axis=0)
    else:
        xs[xs < 0] = 0
        xs[xs > sensor_size[0]] = sensor_size[0]
        ys[ys < 0] = 0
        ys[ys > sensor_size[1]] = sensor_size[1]

    return events



import numpy as np

from .utils import guess_event_ordering_numpy, is_multi_image


def crop_numpy(
    events,
    images=None,
    sensor_size=(346, 260),
    ordering=None,
    target_size=(256, 256),
    multi_image=None,
):
    """Crops the sensor size to a smaller sensor.
    Removes events outsize of the target sensor and maps

    x' = x - new_sensor_start_x

    y' = y - new_sensor_start_y

    Args:
        events: ndarray of shape [num_events, num_event_channels]
        images: ndarray of these possible shapes:
                - [num_images, height, width, num_channels]
                - [height, width, num_channels]
                - [num_images, height, width]
                - [height, width]
        sensor_size: size of the sensor that was used [W,H]
        ordering: ordering of the event tuple inside of events, if None
                 the system will take a guess through
                 guess_event_ordering_numpy. This function requires 'x'
                 and 'y' to be in the ordering
        target_size: size of the sensor that was used [W',H']
        multi_image: Fix whether or not the first dimension of images is
                    num_images

    Returns:
        events - events within the crop box
        images - crop box out of the images
    """

    assert target_size[0] <= sensor_size[0] and target_size[1] <= sensor_size[1]

    if ordering is None:
        ordering = guess_event_ordering_numpy(events)
    assert "x" and "y" in ordering

    if images is not None and multi_image is None:
        multi_image = is_multi_image(images, sensor_size)

    x_start_ind = int(np.random.rand() * (sensor_size[0] - target_size[0]))
    y_start_ind = int(np.random.rand() * (sensor_size[1] - target_size[1]))

    x_end_ind = x_start_ind + target_size[0]
    y_end_ind = y_start_ind + target_size[1]

    if images is not None and multi_image:
        # multiple images NHW or NHWC
        images = images[:, y_start_ind:y_end_ind, x_start_ind:x_end_ind, ...]
    elif images is not None:
        # single image HW or HWC
        images = images[y_start_ind:y_end_ind, x_start_ind:x_end_ind, ...]

    x_loc = ordering.index("x")
    y_loc = ordering.index("y")

    event_mask = (
        (events[:, x_loc] >= x_start_ind)
        * (events[:, x_loc] < x_end_ind)
        * (events[:, y_loc] >= y_start_ind)
        * (events[:, y_loc] < y_end_ind)
    )

    events = events[event_mask, ...]
    events[:, x_loc] -= x_start_ind
    events[:, y_loc] -= y_start_ind

    return events, images



import numpy as np

from .utils import guess_event_ordering_numpy


def mix_ev_streams_numpy(
    events, offsets=None, check_conflicts=False, sensor_size=(346, 260), ordering=None
):

    """Combine two or more event streams into a single stream. Event collisions result in a single spike
    or none if polarities are opposite. Collisions numbering greater than two are handled by consensus.
    While not technically required, it is recommended that all event streams be the same [x,y] dimension.

    Args:
        events: tuple of event streams which are ndarrays of shape [num_events, num_event_channels]
        offsets: tuple of start time offsets for each event stream:
                    - Default all streams start at the same time
                    - Random : applies a random offset from 0 to the timespan of the longest event stream
        check_conflicts: bool, whether or not to check for event collisions. Slower processing if True and probably uneccessary most of the time
        sensor_size: size of the sensor that was used [W,H]
        ordering: ordering of the event tuple inside of events, if None the system
                  will take a guess through guess_event_ordering_numpy.

    Returns:
        a combined event stream
    """

    if ordering is None:
        assert len(events) > 1
        prime_ordering = guess_event_ordering_numpy(events[0])
        orderings = [guess_event_ordering_numpy(x) for x in events]

        for ordering in orderings:
            assert ordering == prime_ordering

        ordering = prime_ordering

    assert "x" and "y" and "t" in ordering

    x_loc = ordering.find("x")
    y_loc = ordering.find("y")
    t_loc = ordering.find("t")
    p_loc = ordering.find("p")

    # shift events to zero
    events = np.array(events)

    if events.dtype != np.object:
        # event streams are all of same size
        events[:, :, t_loc] -= np.tile(events[:, 0, t_loc], (events.shape[1], 1)).T

    if offsets == "Random":
        time_lengths = [x[-1, t_loc] for x in events]
        max_t = np.max(np.array(time_lengths))
        events[:, :, t_loc] -= np.tile(
            np.random.random(events.shape[0]) * max_t, (events.shape[1], 1)
        ).T

    elif offsets is not None:
        events[:, :, t_loc] -= np.tile(np.array(offsets), (events.shape[1], 1)).T

    # Concatenate and sort
    combined_events = np.concatenate(events, axis=0)
    idx = np.argsort(combined_events[:, t_loc])
    combined_events = combined_events[idx]
    num_colisions = 0

    if check_conflicts:
        keep_events = np.ones(len(combined_events))
        for i in range(len(combined_events) - 1):
            ev_1 = combined_events[i]
            ev_1[p_loc] = 0
            ev_2 = combined_events[i + 1]
            ev_2[p_loc] = 0
            if np.array_equal(ev_1, ev_2):
                if np.array_equal(
                    combined_events[i, p_loc], combined_events[i + 1, p_loc]
                ):
                    keep_events[i] = 0
                    i = i + 1
                else:
                    keep_events[i] = 0
                    keep_events[i + 1] = 0
                    i = i + 1
            else:
                keep_events[i] = 1
        num_colisions = len(keep_events) - np.sum(keep_events)
        combined_events = combined_events[keep_events.astype("bool")]

    return combined_events, num_colisions



import numpy as np

from .utils import guess_event_ordering_numpy
from .mix_ev_streams import mix_ev_streams_numpy


def uniform_noise_numpy(
    events,
    sensor_size=(346, 260),
    ordering=None,
    scaling_factor_to_micro_sec=1,
    noise_density=1e-8,
):
    """
    Introduces a fixed number of noise depending on sensor size and noise
    density factor, uniformly distributed across the focal plane and in time.

    Args:
        events: ndarray of shape [num_events, num_event_channels]
        sensor_size: size of the sensor that was used [W,H]
        ordering: ordering of the event tuple inside of events, if None
                  the system will take a guess through
                  guess_event_ordering_numpy. This function requires 'x', 'y'
                  and 'y' to be in the ordering
        scaling_factor_to_micro_sec: this is a scaling factor to get to micro
                                     seconds from the time resolution used in the event stream,
                                     as the noise time resolution is fixed to 1 micro second.
        noise_density: A noise density of 1 will mean one noise event for every
                       pixel of the sensor size for every micro second.

    Returns:
        events + noise events in one array
    """

    if ordering is None:
        ordering = guess_event_ordering_numpy(events)
    assert "x" and "y" and "t" in ordering

    x_index = ordering.find("x")
    y_index = ordering.find("y")
    t_index = ordering.find("t")

    last_timestamp_micro_seconds = events[-1, t_index] * scaling_factor_to_micro_sec
    first_timestamp_micro_seconds = events[0, t_index] * scaling_factor_to_micro_sec

    recording_length_micro_seconds = (
        last_timestamp_micro_seconds - first_timestamp_micro_seconds
    )
    total_number_of_points = recording_length_micro_seconds * np.product(sensor_size)
    number_of_noise_events = int(total_number_of_points * noise_density)

    noise_x = np.random.uniform(0, sensor_size[0], number_of_noise_events).round()
    noise_y = np.random.uniform(0, sensor_size[1], number_of_noise_events).round()
    noise_t = np.random.uniform(
        first_timestamp_micro_seconds,
        last_timestamp_micro_seconds,
        number_of_noise_events,
    )
    noise_p = np.random.choice([-1, 1], number_of_noise_events)

    noise = np.vstack((noise_x, noise_y, noise_t, noise_p)).T
    noise[:, t_index] /= scaling_factor_to_micro_sec

    event_array = (events, noise)

    events, collisions = mix_ev_streams_numpy(
        event_array, sensor_size=sensor_size, ordering=ordering
    )

    return events



import numpy as np

from .utils import guess_event_ordering_numpy


def refractory_period_numpy(
    events, sensor_size=(346, 260), ordering=None, refractory_period=0.5
):
    """Sets a refractory period for each pixel, during which events will be
    ignored/discarded. We keep events if:

        .. math::
            t_n - t_{n-1} > t_{refrac}

    Args:
        events: ndarray of shape [num_events, num_event_channels]
        sensor_size: size of the sensor that was used [W,H]
        ordering: ordering of the event tuple inside of events, if None
                  the system will take a guess through
                  guess_event_ordering_numpy. This function requires 't', 'x'
                  and 'y' to be in the ordering
        refractory_period: refractory period for each pixel in seconds

    Returns:
        filtered set of events.
    """

    if ordering is None:
        ordering = guess_event_ordering_numpy(events)
    assert "t" and "x" and "y" in ordering

    t_index = ordering.find("t")
    x_index = ordering.find("x")
    y_index = ordering.find("y")

    events_copy = np.zeros(events.shape, dtype=events.dtype)
    copy_index = 0
    timestamp_memory = (
        np.zeros((sensor_size[0], sensor_size[1]), dtype=events.dtype)
        - refractory_period
    )

    for event in events:
        time_since_last_spike = (
            event[t_index] - timestamp_memory[int(event[x_index]), int(event[y_index])]
        )
        if time_since_last_spike > refractory_period:
            events_copy[copy_index] = event
            copy_index += 1
        timestamp_memory[int(event[x_index]), int(event[y_index])] = event[t_index]

    return events_copy[:copy_index]



import os
import numpy as np
from torchvision.datasets.vision import VisionDataset
from torchvision.datasets.utils import (
    check_integrity,
    download_and_extract_archive,
    extract_archive,
)


class NCALTECH101(VisionDataset):
    """NCALTECH101 <https://www.garrickorchard.com/datasets/n-caltech101> data set.

    arguments:
        save_to: location to save files to on disk
        download: choose to download data or not
        transform: list of transforms to apply to the data
        target_transform: list of transforms to apply to targets
    """

    url = "https://www.dropbox.com/sh/iuv7o3h2gv6g4vd/AAD0T79lglp-BrdbSCuH_7MFa/Caltech101.zip?dl=1"
    file_md5 = "66201824eabb0239c7ab992480b50ba3"
    filename = "Caltech101.zip"
    folder_name = "Caltech101"

    sensor_size = (233, 173)
    ordering = "xytp"

    def __init__(self, save_to, download=True, transform=None, target_transform=None):
        super(NCALTECH101, self).__init__(
            save_to, transform=transform, target_transform=target_transform
        )

        self.location_on_system = save_to
        self.data = []
        self.targets = []

        if download:
            self.download()

        if not check_integrity(
            os.path.join(self.location_on_system, self.filename), self.file_md5
        ):
            raise RuntimeError(
                "Dataset not found or corrupted."
                + " You can use download=True to download it"
            )

        file_path = self.location_on_system + "/" + self.folder_name
        for path, dirs, files in os.walk(file_path):
            dirs.sort()
            for file in files:
                if file.endswith("bin"):
                    events = self._read_dataset_file(path + "/" + file)
                    self.data.append(events)
                    label_number = os.path.basename(path)
                    self.targets.append(label_number)

    def __getitem__(self, index):
        events, target = self.data[index], self.targets[index]
        if self.transform is not None:
            events = self.transform(events, self.sensor_size, self.ordering)
        if self.target_transform is not None:
            target = self.target_transform(target)
        return events, target

    def __len__(self):
        return len(self.data)

    def download(self):
        download_and_extract_archive(
            self.url, self.location_on_system, filename=self.filename, md5=self.file_md5
        )

    def _read_dataset_file(self, filename):
        f = open(filename, "rb")
        raw_data = np.fromfile(f, dtype=np.uint8)
        f.close()
        raw_data = np.uint32(raw_data)

        all_y = raw_data[1::5]
        all_x = raw_data[0::5]
        all_p = (raw_data[2::5] & 128) >> 7  # bit 7
        all_ts = (
            ((raw_data[2::5] & 127) << 16) | (raw_data[3::5] << 8) | (raw_data[4::5])
        )

        # Process time stamp overflow events
        time_increment = 2 ** 13
        overflow_indices = np.where(all_y == 240)[0]
        for overflow_index in overflow_indices:
            all_ts[overflow_index:] += time_increment

        # Everything else is a proper td spike
        td_indices = np.where(all_y != 240)[0]

        td = np.empty([td_indices.size, 4], dtype=np.int32)
        td[:, 0] = all_x[td_indices]
        td[:, 1] = all_y[td_indices]
        td[:, 2] = all_ts[td_indices]
        td[:, 3] = all_p[td_indices]

        return td



import numpy as np

from .utils import guess_event_ordering_numpy, is_multi_image


def flip_ud_numpy(
    events,
    images=None,
    sensor_size=(346, 260),
    ordering=None,
    flip_probability=0.5,
    multi_image=None,
):
    """
    Flips events and images in y. Pixels map as:

        y' = height - y

    Args:
        events: ndarray of shape [num_events, num_event_channels]
        images: ndarray of these possible shapes:
                [num_images, height, width, num_channels],
                [height, width, num_channels],
                [num_images, height, width],
                [height, width]
        sensor_size: size of the sensor that was used [W,H]
        ordering: ordering of the event tuple inside of events, if None
                  the system will take a guess through
                  guess_event_ordering_numpy. This function requires 'x'
                  to be in the ordering
        flip_probability: probability of performing the flip
        multi_image: Fix whether or not the first dimension of images is num_images

    Returns:
        - returns every event with y' = (sensor_size[1] - 1) - y
        - images flipped in y
    """

    if ordering is None:
        ordering = guess_event_ordering_numpy(events)
    assert "y" in ordering

    if images is not None and multi_image is None:
        multi_image = is_multi_image(images, sensor_size)

    if np.random.rand() <= flip_probability:
        if images is not None and multi_image:
            # multiple images NHW or NHWC
            images = images[:, ::-1, :, ...]
        elif images is not None:
            # single image HW or HWC
            images = images[::-1, :, ...]

        y_loc = ordering.index("y")

        events[:, y_loc] = sensor_size[1] - 1 - events[:, y_loc]

    return events, images



import numpy as np

from .utils import guess_event_ordering_numpy
from .time_skew import time_skew_numpy


def st_transform(
    events,
    spatial_transform,
    temporal_transform,
    sensor_size=(346, 260),
    ordering=None,
    roll=False,
):
    """Transform all events spatial and temporal locations based on
    given spatial transform matrix and temporal transform vector.

    Args:
        events: ndarray of shape [num_events, num_event_channels]
        ordering: ordering of the event tuple inside of events, if None
                  the system will take a guess through
                  guess_event_ordering_numpy. This function requires 't', 'x', 'y'
                  to be in the ordering
        spatial_transform: 3x3 matrix which can be used to perform rigid (translation and rotation),
                           non-rigid (scaling and shearing), and non-affine transformations. Generic to user input.
        temporal_transform: scale time between events and offset temporal location based on 2 member vector.
                            Used as arguments to time_skew method.
        roll: boolean input to determine if transformed events will be translated across sensor boundaries (True).
              Otherwise, events will be clipped at sensor boundaries.
        sensor_size: tuple which stipulates sensor size and used to determine tranform limits.

    Returns:
        the input events with tranformed temporal and spatial location
    """

    if ordering is None:
        ordering = guess_event_ordering_numpy(events)
        assert "x" and "y" and "t" and "p" in ordering

    x_index = ordering.find("x")
    y_index = ordering.find("y")
    t_index = ordering.find("t")
    p_index = ordering.find("p")

    number_events = events.shape[0]

    ones_vec = np.ones((number_events, 1))
    events_homog_coord = np.hstack(
        (
            events[:, x_index].reshape(number_events, 1),
            events[:, y_index].reshape(number_events, 1),
            ones_vec,
        )
    )
    # spatial transform
    events_spatial_transform = np.matmul(spatial_transform, events_homog_coord.T)
    # roll coordinates if specified
    events_spatial_transform_X = events_spatial_transform[x_index, :]
    events_spatial_transform_Y = events_spatial_transform[y_index, :]

    outOfRange_eventsX_P = np.where(
        events_spatial_transform[x_index, :] >= sensor_size[0]
    )  # Out Of Range coordinates based on imDim
    outOfRange_eventsX_N = np.where(events_spatial_transform[x_index, :] < 0)
    outOfRange_eventsY_P = np.where(
        events_spatial_transform[y_index, :] >= sensor_size[1]
    )
    outOfRange_eventsY_N = np.where(events_spatial_transform[y_index, :] < 0)

    if roll:
        events_spatial_transform_X[outOfRange_eventsX_P] -= (
            sensor_size[0] - 1
        )  # roll X right
        events_spatial_transform_X[outOfRange_eventsX_N] += (
            sensor_size[0] - 1
        )  # roll X left
        events_spatial_transform_Y[outOfRange_eventsY_P] -= (
            sensor_size[1] - 1
        )  # roll Y down
        events_spatial_transform_Y[outOfRange_eventsY_N] += (
            sensor_size[1] - 1
        )  # roll Y up
    else:
        events_spatial_transform_X[outOfRange_eventsX_P] = (
            sensor_size[0] - 1
        )  # Clip X pos.
        events_spatial_transform_X[outOfRange_eventsX_N] = 0  # Clip X neg.
        events_spatial_transform_Y[outOfRange_eventsY_P] = (
            sensor_size[1] - 1
        )  # Clip Y pos.
        events_spatial_transform_Y[outOfRange_eventsY_N] = 0  # Clip Y neg.

    tr_events = np.zeros((number_events, 4))
    tr_events[:, x_index] = events_spatial_transform_X.T
    tr_events[:, y_index] = events_spatial_transform_Y.T
    tr_events[:, t_index] = events[:, t_index]
    tr_events[:, p_index] = events[:, p_index]
    tr_events = time_skew_numpy(
        tr_events, ordering, temporal_transform[0], temporal_transform[1]
    )

    return tr_events



import numpy as np
import warnings

from .utils import guess_event_ordering_numpy


def time_skew_numpy(events, ordering=None, coefficient=0.9, offset=0):
    """Skew all event timestamps according to a linear transform,
    potentially sampled from a distribution of acceptable functions.

    Args:
        events: ndarray of shape [num_events, num_event_channels]
        ordering: ordering of the event tuple inside of events, if None
                  the system will take a guess through
                  guess_event_ordering_numpy. This function requires 't'
                  to be in the ordering
        coefficient: a real-valued multiplier applied to the timestamps of the events.
                     E.g. a coefficient of 2.0 will double the effective delay between any
                     pair of events.
        offset: value by which the timestamps will be shifted after multiplication by
                the coefficient. Negative offsets are permissible but may result in
                in an exception if timestamps are shifted below 0.

    Returns:
        the input events with rewritten timestamps (rounded to nearest integer if timestamps used integers in the first place.)
    """

    if ordering is None:
        ordering = guess_event_ordering_numpy(events)
    assert "t" in ordering

    t_index = ordering.index("t")

    if np.issubdtype(events.dtype, np.integer):
        events[:, t_index] = (events[:, t_index] * coefficient + offset).round()
    else:
        events[:, t_index] = events[:, t_index] * coefficient + offset

    assert np.min(events[:, t_index]) >= 0

    return events



import numpy as np


def to_timesurface_numpy(
    events,
    sensor_size,
    ordering,
    surface_dimensions=(7, 7),
    tau=5e3,
    decay="lin",
    merge_polarities=False,
):
    """Representation that creates timesurfaces for each event for one recording.

    Args:
        surface_dimensions (int, int): width does not have to be equal to height, however both numbers have to be odd.
        tau (float): time constant to decay events around occuring event with.
        decay (str): can be either 'lin' or 'exp', corresponding to linear or exponential decay.
        merge_polarities (bool): flag that tells whether polarities should be taken into account separately or not.

    Returns:
        array of timesurfaces with dimensions (w,h)
    """
    radius_x = surface_dimensions[0] // 2
    radius_y = surface_dimensions[1] // 2
    assert "x" and "y" and "t" and "p" in ordering
    assert len(sensor_size) == 2
    x_index = ordering.find("x")
    y_index = ordering.find("y")
    t_index = ordering.find("t")
    p_index = ordering.find("p")
    n_of_events = len(events)
    if merge_polarities:
        events[:, p_index] = np.zeros(n_of_events)
    n_of_pols = len(np.unique(events[:, p_index]))
    timestamp_memory = np.zeros(
        (n_of_pols, sensor_size[0] + radius_x * 2, sensor_size[1] + radius_y * 2)
    )
    timestamp_memory -= tau * 3 + 1
    all_surfaces = np.zeros(
        (n_of_events, n_of_pols, surface_dimensions[0], surface_dimensions[1])
    )
    for index, event in enumerate(events):
        x = int(event[x_index])
        y = int(event[y_index])
        timestamp_memory[int(event[p_index]), x + radius_x, y + radius_y] = event[
            t_index
        ]
        timestamp_context = (
            timestamp_memory[
                :, x : x + surface_dimensions[0], y : y + surface_dimensions[1]
            ]
            - event[t_index]
        )
        if decay == "lin":
            timesurface = timestamp_context / (3 * tau) + 1
            timesurface[timesurface < 0] = 0
        elif decay == "exp":
            timesurface = np.exp(timestamp_context / tau)
            timesurface[timestamp_context < (-3 * tau)] = 0
        all_surfaces[index, :, :, :] = timesurface
    return all_surfaces



import numpy as np


def drop_events_numpy(events, drop_probability=0.5, random_drop_probability=False):
    """Randomly drops events with drop_probability.

    Args:
        events: ndarray of shape [num_events, num_event_channels].
        drop_probability: probability of dropping out event.
        random_drop_probability: randomize the dropout probability
                                 between 0 and drop_probability.

    Returns:
        augmented events that were not dropped.
    """

    if random_drop_probability is True:
        drop_probability *= np.random.rand()

    length = events.shape[0]  # find the number of events
    nDrop = int(drop_probability * length + 0.5)
    ind = np.random.choice(length, nDrop, replace=False)
    return np.delete(events, ind, axis=0)



import numpy as np

from .utils import guess_event_ordering_numpy


def time_jitter_numpy(
    events, ordering=None, variance=1, integer_timestamps=False, clip_negative=True
):
    """Changes timestamp for each event by drawing samples from a
    Gaussian distribution with the following properties:

        mean = [t]
        variance = variance

    Automatically clipping negative timestamps.

    Args:
        events: ndarray of shape [num_events, num_event_channels]
        ordering: ordering of the event tuple inside of events, if None
                  the system will take a guess. This function requires 't'
                  to be in the ordering
        variance: change the variance of the time jitter
        integer_timestamps: will round the jitter that is added to timestamps
        clip_negative: drops events that have negative timestamps, otherwise set to zero.

    Returns:
        temporally jittered set of events.
    """

    if ordering is None:
        ordering = guess_event_ordering_numpy(events)
    assert "t" in ordering

    t_index = ordering.find("t")
    shifts = np.random.normal(0, variance, len(events))

    if integer_timestamps:
        shifts = shifts.round()

    times = events[:, t_index]

    if np.issubdtype(events.dtype, np.integer):
        times += shifts.astype(np.int)
    else:
        times += shifts

    if clip_negative:
        events = np.delete(events, (np.where(times < 0)), axis=0)
    else:
        times[times < 0] = 0

    return events



from .ibmGesture import IBMGesture
from .nmnist import NMNIST
from .ncaltech101 import NCALTECH101
from .ncars import NCARS
from .pokerdvs import POKERDVS
from torch.utils.data import DataLoader

__all__ = ["IBMGesture", "NMNIST", "NCALTECH101", "NCARS", "POKERDVS", "DataLoader"]



import numpy as np
import warnings

from .utils import guess_event_ordering_numpy, is_multi_image


def time_reversal_numpy(
    events,
    images=None,
    sensor_size=(346, 260),
    ordering=None,
    flip_probability=0.5,
    multi_image=None,
):
    """Temporal flip is defined as:

        .. math::
           t_i' = max(t) - t_i

           p_i' = -1 * p_i

    Args:
        events: ndarray of shape [num_events, num_event_channels]
        images: ndarray of these possible shapes:
                [num_images, height, width, num_channels],
                [height, width, num_channels],
                [num_images, height, width],
                [height, width]
        sensor_size: size of the sensor that was used [W,H]
        ordering: ordering of the event tuple inside of events, if None
                  the system will take a guess through
                  guess_event_ordering_numpy. This function requires 'x'
                  to be in the ordering
        flip_probability: probability of performing the flip
        multi_image: Fix whether or not the first dimension of images is num_images

    Returns:
        - every event flipped in time and polarity or no change and
        - images flipped in time
    """

    if ordering is None:
        ordering = guess_event_ordering_numpy(events)
    assert "t" and "p" in ordering

    if images is not None and multi_image is None:
        multi_image = is_multi_image(images, sensor_size)

    if multi_image is not None and not multi_image:
        warnings.warn("Time reversal with single image can lead to incorrect results.")

    if np.random.rand() < flip_probability:
        if images is not None and multi_image:
            # multiple images NHW or NHWC
            images = images[::-1, ...]

        t_loc = ordering.index("t")
        p_loc = ordering.index("p")

        events[:, t_loc] = np.max(events[:, t_loc]) - events[:, t_loc]
        events[:, p_loc] *= -1

    return events, images



import numpy as np

def findCell(x, y, bounds):
    
    # build point
    points = np.array([[x,y]])
    
    # check for each event if all coordinates are in bounds
    allInBounds = (points[:,0] >= bounds[:,None,0])
    allInBounds &= (points[:,1] >= bounds[:,None,1])
    allInBounds &= (points[:,0] < bounds[:,None,2])
    allInBounds &= (points[:,1] < bounds[:,None,3])

    # now find out the positions of all nonzero values
    nz = np.nonzero(allInBounds)

    # initialize the result with all nan
    r = np.full(points.shape[0], np.nan)
    
    # use nz[1] to index event position and nz[0] to tell which cell the event belongs to
    r[nz[1]] = nz[0]
    return int(r)
        
def to_averaged_timesurface(
    events,
    sensor_size,
    ordering,
    cell_size=10,
    surface_size=3,
    temporal_window=5e5,
    tau=5e3,
    decay="lin",
    merge_polarities=False,
):
    """Representation that creates averaged timesurfaces for each event for one recording.

    Args:
        cell_size (int): size of each square in the grid
        surface_size (int): has to be odd
        time_window (float): how far back to look for past events for the time averaging
        tau (float): time constant to decay events around occuring event with.
        decay (str): can be either 'lin' or 'exp', corresponding to linear or exponential decay.
        merge_polarities (bool): flag that tells whether polarities should be taken into account separately or not.

    Returns:
        array of timesurfaces with dimensions (w,h)
    """
    radius = surface_size // 2
    assert surface_size <= cell_size
    assert "x" and "y" and "t" and "p" in ordering
#     assert len(sensor_size) == 1
    x_index = ordering.find("x")
    y_index = ordering.find("y")
    t_index = ordering.find("t")
    p_index = ordering.find("p")
    n_of_events = len(events)
    if merge_polarities:
        events[:, p_index] = np.zeros(n_of_events)
    n_of_pols = len(np.unique(events[:, p_index]))
    
    all_surfaces = np.zeros(
        (n_of_events, n_of_pols, surface_size, surface_size)
    )
    
    # find how many rows and columns we have in the grid
    rows = -(-sensor_size[0] // cell_size)
    cols = -(-sensor_size[1] // cell_size)
    ncells = rows * cols
    
    # initialise cell structures
    cells = np.empty(ncells, dtype=object)
    
    # boundaries for each cell
    xmin = 0;
    ymin = 0;
    bounds = np.zeros((ncells,4))
    for i in np.arange(ncells):
        
        if i != 0 and i % rows == 0:
            xmin=0
            ymin += cell_size
            
        bounds[i] = np.array([xmin,ymin,xmin+cell_size,ymin+cell_size])
        xmin += cell_size
    
    # event loop
    for index, event in enumerate(events):
        x = int(event[x_index])
        y = int(event[y_index])
        
        # find the cell
        r = findCell(x, y, bounds)
        
        # initialise timesurface
        timesurface = np.zeros([surface_size, surface_size])
        timesurface[radius,radius] = 1
            
        if cells[r]:
            local_memory = np.array(cells[r])
            
            # find events ej such that tj is in [ti-temporal_window, ti)
            context  = (local_memory[:,0] >= event[t_index]-temporal_window)
            context &= (local_memory[:,0] < event[t_index])
            
            # find events ej such that xj is in [xi-radius,xi+radius]
            context &= (local_memory[:,1] <= x+radius)
            context &= (local_memory[:,1] >= x-radius) 
            
            # find events ej such that yj is in [yi-radius,yi+radius]
            context &= (local_memory[:,2] <= y+radius)
            context &= (local_memory[:,2] >= y-radius)
            
            # taking into consideration different polarities
            context &= (local_memory[:,3] == event[p_index])
            
            # get the neighborhood of center event
            neighborhood = local_memory[context]
            
            if len(neighborhood) != 0:
                
                # get unique coordinates
                unique_coord = np.unique(neighborhood[:,1:3], axis=0)                
                for i, coord in enumerate(unique_coord):
                    
                    # get timestamp of matching coordinates from cell
                    match  = (neighborhood[:,1] == coord[0])
                    match &= (neighborhood[:,2] == coord[1])
                    
                    # scale x and y to find their position on timesurface
                    scaled_x = int(coord[0] - x + radius)
                    scaled_y = int(coord[1] - y + radius)
                    
                    # for each neighbor use some of decay of past events
                    if decay == "lin":
                        tmp_ts = ((neighborhood[match,0] - event[t_index]) / (3 * tau) + 1)
                        tmp_ts[tmp_ts < 0] = 0
                        timesurface[scaled_x, scaled_y] = np.sum(tmp_ts)
                    elif decay == "exp":
                        timesurface[scaled_x, scaled_y] = np.sum(np.exp((neighborhood[match,0] - event[t_index]) / tau))
                
        else:
            # initialising cell with an empty list
            cells[r] = []
            
        # save event inside the cell
        cells[r].append((event[t_index],x,y,event[p_index]))
        
        # save into all_surfaces
        all_surfaces[index, :, :, :] = timesurface
    
    return all_surfaces



from . import functional


class Compose:
    """Composes several transforms together.
    Args:
        transforms (list of ``Transform`` objects): list of transforms to compose.
    Example:
        >>> transforms.Compose([
        >>>     transforms.Denoise(),
        >>>     transforms.ToTensor(),
        >>> ])
    """

    def __init__(self, transforms):
        self.transforms = transforms

    def __call__(self, events, sensor_size, ordering, images=None, multi_image=None):
        for t in self.transforms:
            events, images = t(
                events=events,
                images=images,
                sensor_size=sensor_size,
                ordering=ordering,
                multi_image=multi_image,
            )
        if multi_image and images.all() != None:
            return events, images
        else:
            return events

    def __repr__(self):
        format_string = self.__class__.__name__ + "("
        for t in self.transforms:
            format_string += "\n"
            format_string += "    {0}".format(t)
        format_string += "\n)"
        return format_string


class Crop:
    """Crops the sensor size to a smaller sensor and removes events outsize of the target sensor and maps."""

    def __init__(self, target_size=(256, 256)):
        self.target_size = target_size

    def __call__(self, events, sensor_size, ordering, images=None, multi_image=None):
        return functional.crop_numpy(
            events, images, sensor_size, ordering, self.target_size, multi_image
        )


class DropEvents:
    """Drops events with  a certain probability."""

    def __init__(self, drop_probability=0.5, random_drop_probability=False):
        self.drop_probability = drop_probability
        self.random_drop_probability = random_drop_probability

    def __call__(self, events, sensor_size, ordering, images=None, multi_image=None):
        events = functional.drop_events_numpy(
            events, self.drop_probability, self.random_drop_probability
        )
        return events, images


class FlipLR:
    """Mirrors x coordinates of events and images (if present)"""

    def __init__(self, flip_probability=0.5):
        self.flip_probability_lr = flip_probability

    def __call__(self, events, sensor_size, ordering, images=None, multi_image=None):
        return functional.flip_lr_numpy(
            events, images, sensor_size, ordering, self.flip_probability_lr, multi_image
        )


class FlipPolarity:
    """Changes polarities 1 to -1 and polarities [-1, 0] to 1"""

    def __init__(self, flip_probability=0.5):
        self.flip_probability_pol = flip_probability

    def __call__(self, events, sensor_size, ordering, images=None, multi_image=None):
        events = functional.flip_polarity_numpy(
            events, self.flip_probability_pol, ordering
        )
        return events, images


class FlipUD:
    """Mirrors y coordinates of events and images (if present)"""

    def __init__(self, flip_probability=0.5):
        self.flip_probability_ud = flip_probability

    def __call__(self, events, sensor_size, ordering, images=None, multi_image=None):
        return functional.flip_ud_numpy(
            events, images, sensor_size, ordering, self.flip_probability_ud, multi_image
        )


class Denoise:
    """Cycles through all events and drops it if there is no other event within
    a time of time_filter and a spatial neighbourhood of 1."""

    def __init__(self, time_filter=10000):
        self.time_filter = time_filter

    def __call__(self, events, sensor_size, ordering, images=None, multi_image=None):
        events = functional.denoise_numpy(
            events, sensor_size, ordering, self.time_filter
        )
        return events, images


class RefractoryPeriod:
    """Cycles through all events and drops event if within refractory period for
    that pixel."""

    def __init__(self, refractory_period=0.5):
        self.refractory_period = refractory_period

    def __call__(self, events, sensor_size, ordering, images=None, multi_image=None):
        events = functional.refractory_period_numpy(
            events, sensor_size, ordering, self.refractory_period
        )
        return events, images


class SpatialJitter:
    """Blurs x and y coordinates of events. Integer or subpixel precision possible."""

    def __init__(
        self,
        variance_x=1,
        variance_y=1,
        sigma_x_y=0,
        integer_coordinates=True,
        clip_outliers=True,
    ):
        self.variance_x = variance_x
        self.variance_y = variance_y
        self.sigma_x_y = sigma_x_y
        self.integer_coordinates = integer_coordinates
        self.clip_outliers = clip_outliers

    def __call__(self, events, sensor_size, ordering, images=None, multi_image=None):
        events = functional.spatial_jitter_numpy(
            events,
            sensor_size,
            ordering,
            self.variance_x,
            self.variance_y,
            self.sigma_x_y,
            self.integer_coordinates,
            self.clip_outliers,
        )
        return events, images


class SpatioTemporalTransform:
    """Choose arbitrary spatial and temporal transformation."""

    def __init__(self, spatial_transform, temporal_transform, roll=False):
        self.spatial_transform = spatial_transform
        self.temporal_transform = temporal_transform
        self.roll = roll

    def __call__(self, events, sensor_size, ordering, images=None, multi_image=None):
        events = functional.st_transform(
            events,
            sensor_size,
            ordering,
            self.spatial_transform,
            self.temporal_transform,
            self.roll,
        )
        return events, images


class TimeJitter:
    """Blurs timestamps of events. Will clip negative timestamps by default."""

    def __init__(self, variance=1, integer_timestamps=False, clip_negative=True):
        self.variance = variance
        self.integer_timestamps = integer_timestamps
        self.clip_negative = clip_negative

    def __call__(self, events, sensor_size, ordering, images=None, multi_image=None):
        events = functional.time_jitter_numpy(
            events, ordering, self.variance, self.integer_timestamps, self.clip_negative
        )
        return events, images


class TimeReversal:
    """Will reverse the timestamps of events with a certain probability."""

    def __init__(self, flip_probability=0.5):
        self.flip_probability_t = flip_probability

    def __call__(self, events, sensor_size, ordering, images=None, multi_image=None):
        return functional.time_reversal_numpy(
            events, images, sensor_size, ordering, self.flip_probability_t, multi_image
        )


class TimeSkew:
    """Scale and/or offset all timestamps."""

    def __init__(self, coefficient=0.9, offset=0):
        self.coefficient = coefficient
        self.offset = offset

    def __call__(self, events, sensor_size, ordering, images=None, multi_image=None):
        events = functional.time_skew_numpy(
            events, ordering, self.coefficient, self.offset
        )
        return events, images


class ToRatecodedFrame:
    """Bin events to frames."""

    def __init__(self, frame_time=5000, merge_polarities=True):
        self.frame_time = frame_time
        self.merge_polarities = merge_polarities

    def __call__(self, events, sensor_size, ordering, images=None, multi_image=None):
        frames = functional.to_ratecoded_frame_numpy(
            events,
            sensor_size,
            ordering,
            frame_time=self.frame_time,
            merge_polarities=self.merge_polarities,
        )
        return frames, images


class ToSparseTensor:
    """Turn event array (N,E) into sparse Tensor (B,T,W,H)."""

    def __init__(self, merge_polarities=False):
        self.merge_polarities = merge_polarities

    def __call__(self, events, sensor_size, ordering, images=None, multi_image=None):
        tensor = functional.to_sparse_tensor_pytorch(
            events, sensor_size, ordering, merge_polarities=self.merge_polarities
        )
        return tensor, images


class ToTimesurface:
    """Create Time surfaces for each event."""

    def __init__(
        self, surface_dimensions=(7, 7), tau=5e3, decay="lin", merge_polarities=False
    ):
        assert len(surface_dimensions) == 2
        assert surface_dimensions[0] % 2 == 1 and surface_dimensions[1] % 2 == 1
        self.surface_dimensions = surface_dimensions
        self.tau = tau
        self.decay = decay
        self.merge_polarities = merge_polarities

    def __call__(self, events, sensor_size, ordering, images=None, multi_image=None):
        surfaces = functional.to_timesurface_numpy(
            events,
            sensor_size,
            ordering,
            self.surface_dimensions,
            self.tau,
            self.decay,
            self.merge_polarities,
        )
        return surfaces, images


class UniformNoise:
    """Inject noise events."""

    def __init__(self, noise_density=1e-8):
        self.noise_density = noise_density

    def __call__(self, events, sensor_size, ordering, images=None, multi_image=None):
        surfaces = functional.uniform_noise_numpy(
            events, sensor_size, ordering, self.noise_density
        )
        return events, images


class Repeat:
    """Copies target n times. Useful to transform sample labels into sequences."""

    def __init__(self, repetitions):
        self.repetitions = repetitions

    def __call__(self, target):
        return np.tile(np.expand_dims(target, 0), [self.repetitions, 1])


class ToOneHotEncoding:
    """Transforms one or more targets into a one hot encoding scheme."""

    def __init__(self, n_classes):
        self.n_classes = n_classes

    def __call__(self, target):
        return np.eye(self.n_classes)[target]


class ToAveragedTimesurface(object):
    """Creates Averaged Time Surfaces."""

    def __init__(
        self,
        cell_size=10,
        surface_size=7,
        temporal_window=5e5,
        tau=5e3,
        decay="lin",
        merge_polarities=False,
    ):
        assert surface_size % 2 == 1
        self.cell_size = cell_size
        self.surface_size = surface_size
        self.temporal_window = temporal_window
        self.tau = tau
        self.decay = decay
        self.merge_polarities = merge_polarities

    def __call__(self, events, sensor_size, ordering, images=None, multi_image=None):
        surfaces = functional.to_averaged_timesurface(
            events,
            sensor_size,
            ordering,
            self.cell_size,
            self.surface_size,
            self.temporal_window,
            self.tau,
            self.decay,
            self.merge_polarities,
        )
        return surfaces, images



import os
import numpy as np
from torchvision.datasets.vision import VisionDataset
from torchvision.datasets.utils import (
    check_integrity,
    download_and_extract_archive,
    extract_archive,
)


class IBMGesture(VisionDataset):
    """IBMGesture <http://research.ibm.com/dvsgesture/> data set.

    arguments:
        save_to: location to save files to on disk
        train: choose training or test set
        download: choose to download data or not
        transform: list of transforms to apply to the data
        target_transform: list of transforms to apply to targets
    """

    # Train: https://www.neuromorphic-vision.com/public/downloads/ibmGestureTrain.tar.gz
    # Test : https://www.neuromorphic-vision.com/public/downloads/ibmGestureTest.tar.gz
    base_url = "https://www.neuromorphic-vision.com/public/downloads/"
    test_zip = base_url + "ibmGestureTest.tar.gz"
    train_zip = base_url + "ibmGestureTrain.tar.gz"
    test_md5 = "56070e45dadaa85fff82e0fbfbc06de5"
    train_md5 = "3a8f0d4120a166bac7591f77409cb105"
    test_filename = "ibmGestureTest.tar.gz"
    train_filename = "ibmGestureTrain.tar.gz"
    classes = [
        "hand_clapping",
        "right_hand_wave",
        "left_hand_wave",
        "right_arm_clockwise",
        "right_arm_counter_clockwise",
        "left_arm_clockwise",
        "left_arm_counter_clockwise",
        "arm_roll",
        "air_drums",
        "air_guitar",
        "other_gestures",
    ]

    sensor_size = (128, 128)
    ordering = "xypt"

    def __init__(
        self, save_to, train=True, download=True, transform=None, target_transform=None
    ):
        super(IBMGesture, self).__init__(
            save_to, transform=transform, target_transform=target_transform
        )
        self.train = train
        self.location_on_system = save_to
        self.data = []
        self.samples = []
        self.targets = []

        if train:
            self.url = self.train_zip
            self.file_md5 = self.train_md5
            self.filename = self.train_filename
            self.folder_name = "ibmGestureTrain"
        else:
            self.url = self.test_zip
            self.file_md5 = self.test_md5
            self.filename = self.test_filename
            self.folder_name = "ibmGestureTest"

        if download:
            self.download()

        if not check_integrity(
            os.path.join(self.location_on_system, self.filename), self.file_md5
        ):
            raise RuntimeError(
                "Dataset not found or corrupted."
                + " You can use download=True to download it"
            )

        file_path = self.location_on_system + "/" + self.folder_name
        for path, dirs, files in os.walk(file_path):
            dirs.sort()
            for file in files:
                if file.endswith("npy"):
                    self.samples.append(path + "/" + file)
                    self.targets.append(int(file[:-4]))

    def download(self):
        download_and_extract_archive(
            self.url, self.location_on_system, filename=self.filename, md5=self.file_md5
        )

    def __getitem__(self, index):
        events = np.load(self.samples[index])
        events[:, 3] *= 1000  # convert from ms to us
        target = self.targets[index]
        if self.transform is not None:
            events = self.transform(events, self.sensor_size, self.ordering)
        if self.target_transform is not None:
            target = self.target_transform(target)
        return events, target

    def __len__(self):
        return len(self.samples)



import numpy as np
import math


def to_ratecoded_frame_numpy(
    events, sensor_size, ordering, frame_time=5000, merge_polarities=True
):
    """Representation that creates frames by encoding the rate of events.

    Args:
        frame_time: time bin for each frame
        merge_polarities: flag to add all polarities together

    Returns:
        numpy array of n rate-coded frames (n,w,h)
    """
    assert "x" and "y" and "t" and "p" in ordering
    assert len(sensor_size) == 2
    x_index = ordering.find("x")
    y_index = ordering.find("y")
    t_index = ordering.find("t")
    p_index = ordering.find("p")
    n_events = len(events)

    pols = events[:, p_index]
    if merge_polarities:
        pols[pols == -1] = 1
        pols[pols == 0] = 1
    else:
        pols[pols == -1] = 0

    n_bins = math.ceil(events[-1, t_index] / frame_time)

    n = 0
    frames = np.zeros((n_bins, sensor_size[1], sensor_size[0]))
    for e in events:
        x = int(e[x_index])
        y = int(e[y_index])
        t = e[t_index]
        p = e[p_index]

        if t > frame_time * (n + 1):
            n += 1
        frames[n, y, x] += p

    frames = np.tanh(frames / 3) * 255

    return frames



import setuptools
import runpy
import os

root = os.path.dirname(os.path.realpath(__file__))
version = runpy.run_path(os.path.join(root, "tonic", "version.py"))["version"]

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="tonic",
    version=version,
    author="The Neuromorphs of Telluride",
    author_email="event-data-augmentation@googlegroups.com",
    description="Spike manipulation and augmentation",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/neuromorphs/tonic",
    include_package_data=False,
    packages=setuptools.find_packages(),
    install_requires=["numpy", "loris", "tqdm"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Intended Audience :: Science/Research",
    ],
)



import numpy as np

from .utils import guess_event_ordering_numpy


def denoise_numpy(events, sensor_size=(346, 260), ordering=None, filter_time=10000):
    """Drops events that are 'not sufficiently connected to other events in the recording.'
    In practise that means that an event is dropped if no other event occured within a spatial neighbourhood
    of 1 pixel and a temporal neighbourhood of filter_time time units. Useful to filter noisy recordings.

    Args:
        events: ndarray of shape [num_events, num_event_channels]
        sensor_size: size of the sensor that was used [W,H]
        ordering: ordering of the event tuple inside of events, if None
                  the system will take a guess through
                  guess_event_ordering_numpy. This function requires 'x',
                  'y' and 't' to be in the ordering
        filter_time: maximum temporal distance to next event, otherwise dropped.
                    Lower values will mean higher constraints, therefore less events.

    Returns:
        filtered set of events.
    """

    if ordering is None:
        ordering = guess_event_ordering_numpy(events)
    assert "x" and "y" and "t" in ordering

    x_index = ordering.find("x")
    y_index = ordering.find("y")
    t_index = ordering.find("t")

    events_copy = np.zeros(events.shape, dtype=events.dtype)
    copy_index = 0
    width = int(sensor_size[0])
    height = int(sensor_size[1])
    timestamp_memory = np.zeros((width, height), dtype=events.dtype) + filter_time

    for event in events:
        x = int(event[x_index])
        y = int(event[y_index])
        t = event[t_index]
        timestamp_memory[x, y] = t + filter_time
        if (
            (x > 0 and timestamp_memory[x - 1, y] > t)
            or (x < width - 1 and timestamp_memory[x + 1, y] > t)
            or (y > 0 and timestamp_memory[x, y - 1] > t)
            or (y < height - 1 and timestamp_memory[x, y + 1] > t)
        ):
            events_copy[copy_index] = event
            copy_index += 1

    return events_copy[:copy_index]



import numpy as np

from .utils import guess_event_ordering_numpy


def flip_polarity_numpy(events, flip_probability=0.5, ordering=None):
    """Flips polarity of individual events with flip_probability.

    Args:
        events: ndarray of shape [num_events, num_event_channels]
        flip_probability: probability of flipping individual event polarities
        ordering: ordering of the event tuple inside of events, if None
                  the system will take a guess through
                  guess_event_ordering_numpy. This function requires 'p'
                  to be in the ordering and requires that polarity is
                  encoded as -1 or 1
    Returns:
        augmented events - returns every event with p' = -p at flip_probability or p' = p at 1 - flip_probability
    """

    if ordering is None:
        ordering = guess_event_ordering_numpy(events)
    assert "p" in ordering

    p_loc = ordering.index("p")

    flips = np.ones(len(events))
    probs = np.random.rand(len(events))
    flips[probs < flip_probability] = -1

    events[:, p_loc] = events[:, p_loc] * flips

    return events



import warnings
import numpy as np


def guess_event_ordering_numpy(events):
    """
    Guesses the names of the channels for events or returns numpy ndarray names

    Arguments:
    - events - the events in [num_events, channels]

    Returns:
    - guess - string representation of ordering of channels
    """

    warnings.warn("[SDAug]::Guessing the ordering of xytp in events")

    if np.issubdtype(events.dtype, np.number):
        if events.shape[1] == 3:
            guess = "xtp"
        elif events.shape[1] == 4:
            guess = "xytp"
        elif events.shape[1] == 5:
            guess = "xyztp"
    elif isinstance(events, (np.ndarray, np.generic)):
        guess = events.dtype.names
    else:
        raise NotImplementedError("Unable to guess event ordering")

    warnings.warn("[SDAug]::Guessed [%s] as ordering of events" % guess)

    return guess


def is_multi_image(images, sensor_size):
    """
    Guesses at if there are multiple images inside of images

    Arguments:
    - images - image array to find where sensor_size is supported shapes
               include
               - [num_images, height, width, num_channels]
               - [height, width, num_channels]
               - [num_images, height, width]
               - [height, width]
    - sensor_size - sensor [W,H]

    Returns:
    - guess - best guess at if there are multiple images
    """

    warnings.warn("[SDAug]::Guessing if there are multiple images")
    if len(images.shape) == 4:
        guess = True
    elif len(images.shape) == 3:
        if images.shape[0] == sensor_size[0]:
            guess = False  # HWC
        else:
            guess = True  # NHW
    elif len(images.shape) == 2:
        guess = False
    else:
        raise NotImplementedError()
    warnings.warn("[SDAug]::Guessed [%s]" % str(guess))

    return guess



import os
import numpy as np
from torchvision.datasets.vision import VisionDataset
from torchvision.datasets.utils import (
    check_integrity,
    download_and_extract_archive,
    extract_archive,
)
from numpy.lib.recfunctions import structured_to_unstructured
import loris


class NCARS(VisionDataset):
    """N-Cars <https://www.prophesee.ai/dataset-n-cars-download/> data set.

    arguments:
        save_to: location to save files to on disk
        train: choose training or test set
        download: choose to download data or not
        transform: list of transforms to apply to the data
        target_transform: list of transforms to apply to targets
    """

    # Train: https://www.neuromorphic-vision.com/public/downloads/ibmGestureTrain.tar.gz
    # Test : https://www.neuromorphic-vision.com/public/downloads/ibmGestureTest.tar.gz
    url = "http://www.prophesee.ai/resources/Prophesee_Dataset_n_cars.zip"
    filename = "Prophesee_Dataset_n_cars.zip"
    train_file = "n-cars_train.zip"
    test_file = "n-cars_test.zip"
    file_md5 = "553ce464d6e5e617b3c21ce27c19368e"
    train_md5 = "976d126a651b95d81800b05a3093337b"
    test_md5 = "3b5e8e9a5bffeb95614b8c0a2ba4e511"
    classes = ["background", "car"]

    class_dict = {"background": 0, "cars": 1}

    sensor_size = (304, 240)
    ordering = "txyp"

    def __init__(
        self, save_to, train=True, download=True, transform=None, target_transform=None
    ):
        super(NCARS, self).__init__(
            save_to, transform=transform, target_transform=target_transform
        )

        self.location_on_system = save_to
        self.data = []
        self.targets = []

        if download:
            self.download()

        if not check_integrity(
            os.path.join(self.location_on_system, self.filename), self.file_md5
        ):
            raise RuntimeError(
                "Dataset not found or corrupted."
                + " You can use download=True to download it"
            )

        if train:
            target_zip = self.train_file
            source_path = os.path.join(save_to, "train")
            target_path = os.path.join(save_to, "ncars-train")
        else:
            target_zip = self.test_file
            source_path = os.path.join(save_to, "test")
            target_path = os.path.join(save_to, "ncars-test")

        if not os.path.exists(target_path):
            self.extract_archive(os.path.join(save_to, target_zip))
            os.rename(source_path, target_path)

        # We will not be loading everything into memory. Instead, we will keep a list of samples into file
        # Could have reused self.data for that purpose as well.
        self.samples = []

        file_path = target_path
        for path, dirs, files in os.walk(file_path):
            dirs.sort()
            for file in files:
                if file.endswith("dat"):
                    self.samples.append(path + "/" + file)
                    self.targets.append(self.class_dict[os.path.basename(path)])

    def __getitem__(self, index):
        events = loris.read_file(self.samples[index])["events"]
        events = np.array(structured_to_unstructured(events))
        target = self.targets[index]
        if self.transform is not None:
            events = self.transform(events, self.sensor_size, self.ordering)
        if self.target_transform is not None:
            target = self.target_transform(target)
        return events, target

    def __len__(self):
        return len(self.samples)

    def download(self):
        download_and_extract_archive(
            self.url, self.location_on_system, filename=self.filename, md5=self.file_md5
        )
